# -*- coding: iso-8859-15 -*-
# MIPS analyser for IDA
#
# MIPS (MIPS IDA PluginS)
#
# by Raphal Rigo
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

"""
Documentation :
	This plugin analyses position-independant MIPS (both big and little endian) ELF
	(32bit) files produced by GNU tools (gcc and binutils) (maybe others, if they
	follow the SVR4 PIC convention)

	It does the following :
		- analyses references to the GOT table and add the addresses accessed by the
			lw reg, disp($gp)
		instruction as a comment (or as a manual instruction). It also adds xrefs.
		- if those addresses have a meaningful name (set in IDA), this name is used
		- if those addresses are a stub used to call an external function, the function
		name is loaded from the dynamic symbols and used
		- if the address is a base address (loaded from the beginning of the GOT), it looks
		for the instruction adding the displacement and does the analysis on the
		reconstructed address
		- analyses 'switch' statements implemented with jump tables

        References:
		http://cr0.org/paper/mips.elf.external.resolution.txt
		http://www.caldera.com/developers/devspecs/mipsabi.pdf (System V ABI)

	Usage :
	        If you want, use ident_func.py to identify a maxmimum of functions.

		Go into the function you want to analyse, load the script.
		If you want to analyse the whole segment, select YES when asked :)

		If it asks for the size of the local got entries, you should check yourself if
		the given size is OK before accepting (or you can blindly trust the script)
		by checking the first entries of the GOT.

		Since the functions are marked when using the "whole segment" option, you can
		do several passes to maximize the number of analyzed functions (I advise
		using ident_func.py before running the script rather than running this one
		several times)

		The mark is ignored if you choose to analyse the current function only,
		allowing you to reanalyse it to take advantage of any updates to the db.

	Options:
		- You can activate debug messages by setting the "debug" variable to a non zero int
		- You can choose to set the results of the analysis as comments and/or manual insn
		- You can choose whether to search for the "gcc -O0" case or not with the
		check_for_O0 variable (activated by default)

	Known Problems :
		- if the referred name is displayed in hex, and afterward the corresponding offset
		  gets a name, the hex representation won't be updated
		- if a base address is loaded in a register $reg, and the following happens:
			- an inconditionnal jump is present before the register is accessed
			- the register is accessed after more than X instructions (defaults to 10)
			- the register is accessed several times after being loaded once
		then the reference is not calculated

	TODO :
		- Allow automatic usage from command line
		- Block based analysis with register tracking (for offset + displ case)

        History:
		1.5.3 (February 2009)
			Bugfix : forgot some check in switch tables parsing
		1.5.2 (September 2008)
			Bugfix : some offsets were parsed as positive
		1.5.1 (October 2007)
			Bugfix : detect dummy names correctly
		1.5 (December 2006)
		        - Calls IDA's analysis to help avoid hex references
			- Uses manual operands instead of manual instructions to help
			  track name changes (thanks Ilfak)
		1.4 (September 2006)
			Added normal binding handling (NULL GOT entries)
		1.3 (August 2006)
			Added function marking to avoid second analysis
		1.2 (April 2006)
			Added static files support
			Added weird case (gcc -O0 ?) support
		1.1 (April 2006)
			Added support for switch tables
		1.0 (March 2006)
			Initial Release
"""

import re
import struct
# ELF reader
import Elf
from elf_lookup import *
from idaapi import *

registers = {"$0": 0, "$at":1, "$v0":2, "$v1":3,"$a0":4,"$a1":5,"$a2":6,"$a3":7,
     "$t0":8,"$t1":9,"$t2":10,"$t3":11,"$t4":12,"$t5":13,"$t6":14,"$t7":15,
     "$s0":16,"$s1":17,"$s2":18,"$s3":19,"$s4":20,"$s5":21,"$s6":22,"$s7":23,
     "$t8":24,"$t9":25,"$k0":26,"$k1":27,"$gp":28,"$sp":29,"$fp":30,"$ra":31}

# Set to 1 if you want the instructions to be modified
set_manual_insn = 1

# Set to 1 if you want to set comments
set_comments = 0

# make debug non zero to have debug messages
debug = 0

# Check for special gcc -O0 case
check_for_O0 = 1

# Global vars
gp = 0
got = 0
local_got_end = 0
max_insn_dist = 20

# This function checks is the given ea is a stub to call an external function
# is this is the case, it returns the index of the function called
# else it returns False
# 10 80 99 8F                 lw      $t9, -0x7FF0($gp)
# 21 78 E0 03                 move    $t7, $ra
# 09 F8 20 03                 jalr    $t9
# XX XX 18 24                 li      $t8, XXXX
def isExtStub(ea):
    dword1 = 0x8F998010L
    dword2 = 0x03e07821
    dword3 = 0x0320f809
    word4 = 0x2418
    
    if little_endian==1:
        word_offset = 14
        value_offset = 12
    else:
        # on big endian archs, the last instruction is 24 18 XX XX
        word_offset = 12
        value_offset = 14

    if (Dword(ea)==dword1) & (Dword(ea+4)==dword2) & (Dword(ea+8)==dword3) & (Word(ea+word_offset)==word4):
        return Word(ea+value_offset)
    else:
        return False

# analyse the given ea and returns an insn_t object
def GetOperand(ea, op):
    ua_ana0(ea)
    return get_instruction_operand(get_current_instruction(), op)

# This function looks for the next instruction referring to a register
# storing an offset
# possibilities :
# 1) addiu $reg, imm
# 2) addiu $reg2, $reg, imm
# 3) xxx $regX, off($reg)
# TODO : if xxx $reg, yyy with xxx != addiu & yyy != imm => BADADDR
def findNextInsn(ea,  register, nb):
    cur_i = ea
    end_ea = ea+4*nb
    seen_jump = False
    
    while cur_i<=end_ea:
        if GetMnem(cur_i)==("b"):
            seen_jump = True
            cur_i = cur_i+4
        
        op1 = GetOperand(cur_i, 0)            
        op2 = GetOperand(cur_i, 1)

        # Case 1)
        if (GetMnem(cur_i).find("add")!=-1) & (op1.reg==register) & (op2.type==5):
            return [cur_i, struct.unpack("h",struct.pack("H",GetOperand(cur_i, 1).value&0xFFFF))[0]]

        op3 = GetOperand(cur_i, 2)
        # Case 2
        if (GetMnem(cur_i).find("add")!=-1) & (op2.reg==register) & (op3.type==5):
            return [cur_i, struct.unpack("h",struct.pack("H",GetOperand(cur_i, 2).value&0xFFFF))[0]]
        # Case 3
        if (op2.reg==register) & (op2.type==4):
            return [cur_i, struct.unpack("h",struct.pack("H",op2.addr&0xFFFF))[0]]

        if seen_jump:
            return BADADDR

        cur_i = cur_i + 4
    return BADADDR

# Print debug message
def debug_msg(string):
    if debug:
        print string
    return

# Raises an exception displaying the message
def error_msg(string):
    raise string


# This function analyses switch statements using a jump table
# the code must have the following form :
# 24 63 FF 9D                 addiu   $v1, -"c"
# 2C 62 00 10                 sltiu   $v0, $v1, 16    # NUMBER OF ENTRIES
# 10 40 00 43                 beqz    $v0, loc_40D290
# 8F 86 80 18                 lw      $a2, offset 0x10000000
# 00 03 10 80                 sll     $v0, $v1, 2
# 8F 83 80 1C                 lw      $v1, offset 0x40ec20
# 24 63 EC 20                 addiu   $v1, 0xEC20    # ADD DISPLACEMENT (Xref here)
# 00 43 10 21                 addu    $v0, $v1
# 8C 42 00 00                 lw      $v0, 0($v0)
# 00 5C 10 21                 addu    $v0, $gp
# 00 40 00 08                 jr      $v0            # Jump to entry
def analyseSwitch(ea):
	debug_msg("Analyzing switch at : "+hex(ea))
        # Look for the Xref to the jump table
        xref = Dfirst(ea)
        count = 0
        while (xref == BADADDR) & (count < max_insn_dist):
            ea = ea-4
            count = count + 1
            if GetMnem(ea) == 'addiu':
                xref = Dfirst(ea)

	if (xref == BADADDR):
	    debug_msg("Failed to find table")
	    return

	debug_msg("Found xref to table "+hex(xref)+" at : "+hex(ea))
        # Look for the number of elements in the table
        count = 0
        while (GetMnem(ea) != 'sltiu') & (count < max_insn_dist):
            ea = ea-4
            count = count + 1
	
	if (count >= max_insn_dist):
	    debug_msg("Failed to find number of elements")
	    return
            
        size = GetOperand(ea,2).value
	debug_msg("Found switch size ("+hex(size)+") at :"+hex(ea))
        for j_table_entry in range(xref, xref+4*size, 4):
            entry = Dword(j_table_entry)
            entry = (entry + gp)&0xFFFFFFFF
            # Add references
            add_dref(j_table_entry, entry, dr_O)
            if set_comments:
                MakeComm(j_table_entry, hex(entry))
                         
            if set_manual_insn:
                PatchDword(j_table_entry, entry)
                OpOff(j_table_entry, -1, 0)
                
	return

# Check if we have an instruction probably referring to the GOT
def isInterestingInstruction(ea):
	if check_for_O0:
		# The shitty version of gcc -O0 ?
		# 2 instructions waste !
		# lui $reg, 0
		# addu $reg, $gp
		# lw $reg, -0xFFFF($reg)
		if (GetOperand(ea,1).reg==registers['$gp']) & (GetMnem(ea)=='addu'):
			if (GetOperand(ea+4,1).reg==GetOperand(ea,0).reg) & (GetOpType(ea+4,1)==4):
				return ea+4
	
	# If $gp is used as a base adress in the right operand
	# ie : lw $reg, -0xFFFF($gp)
	if (GetOperand(ea,1).reg==registers['$gp']) & (GetOpType(ea, 1)==4):
		return ea
	return BADADDR

# Check for a jr $reg instruction, where $reg != ($ra | $t9)
# which is most probably the sign of a switch statement
def isJumpSwitchTable(ea):
	if check_for_O0:
		# Once again the shitty version
		# lw $v1, 0($v1)
		# add $v1, $gp
		# jr $v1
		if (GetMnem(ea) == "jr") & (GetOperand(ea,0).reg != registers['$t9']) & (GetOperand(ea,0).reg != registers['$ra']):
			if (GetMnem(ea-4)=='addu')&(GetOperand(ea-4,1).reg==registers['$gp']):
				return True

	# Normal jr $v0 version
	if (GetMnem(ea) == "jr") & (GetOperand(ea,0).reg == registers['$v0']):
		return True

	return False

# Analyse a function
def analyseFunction(elf, fname, fstart, fend):
    print "We're in "+fname+" starting at : "+hex(fstart)+" ending at : "+hex(fend)
    # For all instructions
    for loop_i in range(fstart, fend, 4):
		# check for reference to the got section
		# isInterestingInstruction returns the instruction to analyse or BADADDR
        cur_i = isInterestingInstruction(loop_i)
        if (cur_i != BADADDR):
           debug_msg("----------- "+hex(cur_i)+" -----------")
           offset = struct.unpack("h",struct.pack("H",GetOperand(cur_i, 1).addr&0xFFFF))[0]
	   debug_msg("offset : "+str(offset))
           ref = gp+offset
           debug_msg("'"+GetDisasm(cur_i)+"' offset : "+hex(offset))
           debug_msg("Real address : "+hex(ref))

           # Add a data XREF (as a read access) from the current instruction to
           # the GOT entry
           add_dref(cur_i, ref, dr_R)

           # Get the value pointed by offset($gp), i.e. the GOT entry
           pointed = Dword(ref)

           # The instruction address used for Xrefs
           real_insn = cur_i

           # Handle the case where it loads some base offset (text addressing base ?)
           # then adds some displacement
           if (ref < got+local_got_end):
               debug_msg("Base offset : "+hex(pointed))
               # look for the instruction where the displacement is added
               next_insn = findNextInsn(cur_i, GetOperand(cur_i, 0).reg, max_insn_dist)
               if next_insn!=BADADDR:
                   real_insn = next_insn[0]
                   debug_msg("Next insn: "+hex(real_insn) + " Displacement : "+hex(next_insn[1]))
                   pointed = pointed+next_insn[1]               
           
	   # If the register is $t9, it will probably be called
           if GetOperand(real_insn,0).reg==registers['$t9']:
               AddCodeXref(real_insn, pointed, fl_CF)
           else:
               add_dref(real_insn, pointed, dr_O)               
           
           # Check if we have a meaningful name (i.e. not autogenerated (loc_ and unk_) or
           # previously set by us (_stub)
           name = GetTrueNameEx(cur_i, pointed)

	   # if we have no name, maybe it can have one but IDA didn't analyze it
	   # so force analysis
           if (name == "") & (isLoaded(pointed)):
		analyze_area(pointed, pointed+1)
		name = GetTrueNameEx(cur_i, pointed)
	
	   debug_msg("Name for "+hex(pointed)+" : "+name);

           if (name == "") | has_dummy_name(getFlags(pointed)) | (name.find("_stub")!=-1):            
               # No name => default to address in hex
               if name.find("sub_") != 0:
	         name = hex(pointed)

               # Is the address valid in IDA ?
               valid_addr = isLoaded(pointed);
               debug_msg(name + " isLoaded ?: " + str(valid_addr))
               if (valid_addr == False):
			if pointed == 0:
			   index = (ref-(got+local_got_end))/4+elf.mips_gotsym
			   name = elf.dynamic_symbols[index].name
			   debug_msg(hex(ref)+ ' normal binding, index: ' + str(index)+ ' symbol : '+name);
			  
			   # Add an xref to the symbol table for easier reading
			   add_dref(real_insn, LocByName(name), dr_R)
                           comment = hex(pointed)+" : "+name+" (dyn)"
			else:
			   comment = name+ " (invalid)"
               else:
                   # Check if this is a stub to call an external function
                   extFuncIndex = isExtStub(pointed);
                   debug_msg("External : "+str(extFuncIndex))
                   if extFuncIndex != False:
                       name = elf.dynamic_symbols[extFuncIndex].name
                       # Add an xref to the symbol table for easier reading
                       add_dref(real_insn, LocByName(name), dr_R)
                       MakeNameEx(pointed, name+"_stub",SN_AUTO)
                       comment = hex(pointed)+" : "+name+" (dyn)"
                   else:
                       # Valid address without any name
                       comment = hex(pointed)
                       
                   # Since the address is valid, we can make the GOT entry an offset
                   OpOff(ref, -1, 0)
           else:
               valid_addr = True
               # We have a meaningful name directly from IDA
               comment = hex(pointed)+" : "+name
                 
               # Since we have a name, we can safely make the GOT entry an offset
               OpOff(ref, -1, 0)
               
               
           # Add the name as a comment
           if set_comments:
               MakeComm(cur_i, comment)

           # Put the name in the instruction
           if set_manual_insn:
               if valid_addr:
	       	   name = "offset "+name
	       set_forced_operand(cur_i, 1, name)
           debug_msg("--------------------------------")

        # Search for jump tables
        elif isJumpSwitchTable(loop_i):
            analyseSwitch(loop_i)

    return


def main():
		# Global vars
		global little_endian
		global gp
		global got
		global local_got_end

		# We only support target mips
		proc_module = get_idp_name()
		if proc_module != "mips":
			error_msg("Error, this script only supports MIPS files")

		# First we load the ELF file
		try:
			elf = Elf.Elf(GetInputFilePath())
		except:
			error_msg("Error loading Elf file : "+GetInputFilePath())

		if (elf.e_flags&EF_MIPS_CPIC==0)&(elf.e_flags&EF_MIPS_PIC==0):
			error_msg("Not an SVR4-PIC ELF File !")

		if (ord(elf.e_ident[EI_DATA])==ELFDATA2LSB):
			print "Byte order : Little Endian"
			little_endian = 1
		else:
			print "Byte order : Big Endian"
			little_endian = 0


		# Get GOT address from the elf file
		got = elf.pltgot
		# if this is 0, we have a static file, fallback on IDA
		# and ask for the local_got_end
		if got == 0:
			got = SegByName('.got')
			# Heuristic to determine the local entries
			n = 0xC
			while (Dword(got+n)!=0) and ((Dword(got+n)&0xFFFF)==0):
				n=n+4
			local_got_end = AskLong(n-4, "Please enter the size of the local GOT entries")
			if not local_got_end:
				return
		else:
			local_got_end = 4*elf.mips_local_gotno

		# Local entries in the GOT table
		# We consider that all values loaded from there are used like this :
		# lw $reg, off($gp)
		# addiu $reg, displacement (16bits, signed)
		debug_msg("Size of local GOT entries :"+hex(local_got_end))
			
		# Get GOT address from IDA's segment table
		if (got == BADADDR) or (got == 0):
			error_msg("Cannot find '.got' segment in IDA")

		# $gp is always equal to this
		gp = got + 0x7FF0

		debug_msg("GOT : "+hex(got)+" / $gp : "+hex(gp))

		do_whole_segment = AskYN(0, "Do you want to analyse the whole segment ?")

		if do_whole_segment==1:
			for func_ea in Functions(SegStart(ScreenEA()), SegEnd(ScreenEA())):
				if Comment(func_ea) != 'mips_analyser':
					analyseFunction(elf, GetFunctionName(func_ea), func_ea, FindFuncEnd(func_ea))
					# add comment
					MakeComm(func_ea, "mips_analyser");
		elif do_whole_segment==0:
			fname = GetFunctionName(ScreenEA())
			fstart = LocByName(fname)
			fend = FindFuncEnd(fstart)
			analyseFunction(elf, fname, fstart, fend)

		return

main()
